use microrm::prelude::*;
use microrm::schema::datum::{Datum, DatumDiscriminatorRef};
use microrm::schema::entity::{Entity, EntityID, EntityPartList, EntityPartVisitor};

use crate::schema::UIDCDatabase;
use crate::UIDCError;

mod clap_interface;

pub use clap_interface::ClapInterface;

mod role;
mod user;

pub trait Object: Sized + Entity + std::fmt::Debug {
    type CreateParameters: clap::Parser + std::fmt::Debug;
    fn create_from_params(_: &Self::CreateParameters) -> Result<Self, UIDCError>;
    fn extra_commands() -> impl Iterator<Item = clap::Command> {
        vec![].into_iter()
    }

    /// get the relevant IDMap from the database
    fn db_object(db: &UIDCDatabase) -> &IDMap<Self>
    where
        Self: Sized;

    fn shortname(&self) -> &str;
}

/*
pub trait ObjectExt: Sized + Object {
    fn attach<L: Entity, R: Entity>(
        ctx: impl Queryable<EntityOutput = L>,
        local_uniques: <L::Uniques as EntityPartList>::DatumList,
        relation: String,
        remote_uniques: <R::Uniques as EntityPartList>::DatumList,
    ) -> Result<(), UIDCError> {
        ctx.count();
        // ctx.unique(Self::build_uniques(&local_uniques));
        // ctx.connect_to

        Ok(())
    }

    fn create(ctx: &impl Insertable<Self>, cp: &Self::CreateParameters) -> Result<(), UIDCError> {
        ctx.insert(Self::create_from_params(cp)?)?;
        Ok(())
    }

    fn delete(
        ctx: impl microrm::prelude::Queryable<EntityOutput = Self>,
        which: <Self::Uniques as EntityPartList>::DatumList,
    ) -> Result<(), UIDCError> {
        ctx.unique(which).delete()?;
        Ok(())
    }

    fn list_all(
        ctx: impl microrm::prelude::Queryable<EntityOutput = Self>,
    ) -> Result<(), UIDCError> {
        println!(
            "Listing all {}(s): ({})",
            Self::entity_name(),
            ctx.clone().count()?
        );
        for obj in ctx.get()?.into_iter() {
            println!(" - {}", obj.shortname());
        }

        Ok(())
    }

    fn inspect(
        ctx: impl microrm::prelude::Queryable<EntityOutput = Self>,
        which: <Self::Uniques as EntityPartList>::DatumList,
    ) -> Result<(), UIDCError> {
        let obj = ctx
            .unique(which)
            .get()?
            .ok_or(UIDCError::Abort("no such element, cannot inspect"))?;
        println!("{:#?}", obj.as_ref());

        fn inspect_ai<AI: AssocInterface>(name: &'static str, ai: &AI) {
            println!("{}: ({})", name, ai.count().unwrap());
            for a in ai.get().expect("couldn't get object associations") {
                println!("[#{:3}]: {:?}", a.id().into_raw(), a.wrapped());
            }
        }

        struct AssocFieldWalker;
        impl EntityPartVisitor for AssocFieldWalker {
            fn visit_datum<EP: microrm::schema::entity::EntityPart>(&mut self, datum: &EP::Datum) {
                struct Discriminator<'l, D: Datum>(&'l D, &'static str);

                impl<'l, D: Datum> DatumDiscriminatorRef for Discriminator<'l, D> {
                    fn visit_serialized<T: serde::Serialize + serde::de::DeserializeOwned>(
                        &mut self,
                        _: &T,
                    ) {
                    }
                    fn visit_bare_field<T: Datum>(&mut self, _: &T) {}
                    fn visit_entity_id<E: Entity>(&mut self, _: &E::ID) {}
                    fn visit_assoc_map<E: Entity>(&mut self, amap: &AssocMap<E>) {
                        inspect_ai(self.1, amap);
                    }
                    fn visit_assoc_domain<R: microrm::schema::Relation>(
                        &mut self,
                        adomain: &microrm::schema::AssocDomain<R>,
                    ) {
                        inspect_ai(self.1, adomain);
                    }
                    fn visit_assoc_range<R: microrm::schema::Relation>(
                        &mut self,
                        arange: &microrm::schema::AssocRange<R>,
                    ) {
                        inspect_ai(self.1, arange);
                    }
                }

                datum.accept_discriminator_ref(&mut Discriminator(datum, EP::part_name()));
            }
        }

        obj.accept_part_visitor_ref(&mut AssocFieldWalker);

        Ok(())
    }
}

impl<T: Object + std::fmt::Debug> ObjectExt for T {}
*/
